/*
  Example command line program to demonstrate usage of the Ayyi Auditioner
  service.

  --------------------------------------------------------------------

  This file is part of the Ayyi project. http://ayyi.org
  copyright (C) 2011-2015 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>
#include <math.h>
#include <sys/time.h>
#include <glib.h>
#include <dbus/dbus-glib-bindings.h>
#ifdef USE_SYSLOG
#  include <syslog.h>
#endif
#include "utils.h"

#define APPLICATION_SERVICE_NAME "org.ayyi.Auditioner.Daemon"
#define DBUS_APP_PATH            "/org/ayyi/auditioner/daemon"
#define DBUS_INTERFACE           "org.ayyi.auditioner.Daemon"

static void show_status ();
static void log_handler (const gchar* log_domain, GLogLevelFlags, const gchar*, gpointer);

static const struct option long_options[] = {
	{ "queue",            0, NULL, 'q' },
	{ "status",           0, NULL, 's' },
	{ "verbose",          1, NULL, 'v' },
	{ "version",          0, NULL, 'V' },
	{ "help",             0, NULL, 'h' },
};

static const char* const short_options = "qsv:Vh";

static const char* const usage =
	"Ayyi Audition " PACKAGE_VERSION " copyright Tim Orford 2011-2015.\n\n"
	"Usage: %s [ options ] sample\n"
	" -q --queue     delay playback until all other samples have finished playing.\n"
	" -s --status    show server status information.\n"
	" -v --verbose   show more information.\n"
	" -V --version   print version and quit.\n"
	" -h --help      show this usage information and quit.\n\n"
	"Ayyi Audition is an example client for use with the Ayyi Auditioner service for auditioning audio samples.\n"
	"The org.ayyi.Auditioner.Daemon service is started on-demand.\n"
    "Available commands are\n    StartPlayback\n    StopPlayback\n    QueueForPlayback\n    getStatus\n"
	"Emitted signals are\n    PlaybackStopped\n    Position"
	"\n";

extern int debug;

typedef enum
{
	COMMAND_NONE,
	COMMAND_STATUS,
	COMMAND_QUEUE,
} Command;

int
main(int argc, char** argv)
{
	prefer_stdout = true;
	debug = 0;
	Command command = COMMAND_NONE;

	g_log_set_handler (NULL, G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gtk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("Gdk", G_LOG_LEVEL_CRITICAL | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);
	g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);

#ifndef HAVE_GLIB_2_32
	g_type_init();
	g_thread_init(NULL);
#endif

	int opt;
	while((opt = getopt_long (argc, argv, short_options, long_options, NULL)) != -1) {
		switch(opt) {
			case 'q':
				printf("using queue option\n");
				command = COMMAND_QUEUE;
				break;
			case 's':
				printf("getting server status...\n");
				command = COMMAND_STATUS;
				break;
			case 'v':
				printf("using debug level: %s\n", optarg);
				int d = atoi(optarg);
				if(d<0 || d>5) { pwarn ("bad arg. debug=%i", d); } else debug = d;
				break;
			case 'V':
				printf ("%s %s\n\n", basename(argv[0]), PACKAGE_VERSION);
				printf(
					"Copyright (C) 2011-2015 Tim Orford\n"
					"This is free software; see the source for copying conditions.  There is NO\n"
					"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n"
					);
				return EXIT_SUCCESS;
				break;
			case 'h':
				printf(usage, argv[0]);
				return EXIT_SUCCESS;
				break;
			default:
				break;
		}
	}

#if 0
#ifdef USE_SYSLOG //just testing...
	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("Ayyi.Audition", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);
	syslog (LOG_NOTICE, "Program started by User %d", getuid ());
	syslog (LOG_INFO, "A tree falls in a forest");
	closelog ();
#endif
#endif

	gboolean auditioner_connect(DBusGProxy** proxy)
	{
		p(1, "connecting...");
		GError* error = NULL;
		DBusGConnection* connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
		if(!connection){
			errprintf("failed to get dbus connection\n");
			return FALSE;
		}
		if(!(*proxy = dbus_g_proxy_new_for_name (connection, APPLICATION_SERVICE_NAME, DBUS_APP_PATH, DBUS_INTERFACE))){
			errprintf("failed to get Auditioner\n");
			return FALSE;
		}

		return TRUE;
	}

	DBusGProxy* auditioner;
	if(!auditioner_connect(&auditioner)) return EXIT_FAILURE;

	switch (command){
		case COMMAND_STATUS:
			show_status(auditioner);
			return EXIT_SUCCESS;
			break;
		default:
			break;
	}

	char* sample_path = NULL;
	if (optind < argc) {
		sample_path = argv[optind++];
	}
	if(!sample_path){
		printf("no sample specified\n\n");
		printf(usage, argv[0]);
		return EXIT_FAILURE;
	}
	printf("playing: %s\n", sample_path);

	if(sample_path[0] != '/'){
		char* c = g_get_current_dir();
		sample_path = g_build_path("/", c, sample_path, NULL);
		g_free(c);
	}

#if 0
	if(!dbus_g_proxy_call(proxy, "StartPlayback", &error, G_TYPE_STRING, "filename", G_TYPE_INVALID)){
		if(error){
			dbg(0, "error: %s", error->message);
			g_error_free(error);
		}
		return FALSE;
	}
	dbg(0, "call ok");
#else
	p(1, "calling startPlayback...");
	dbus_g_proxy_call_no_reply(auditioner, command == COMMAND_QUEUE ? "QueueForPlayback" : "StartPlayback", G_TYPE_STRING, sample_path, G_TYPE_INVALID);
#endif

#if 0
	static GMainLoop* loop = NULL;
	g_main_loop_run (loop = g_main_loop_new (NULL, 0));
#endif
	//sleep(2);

	return EXIT_SUCCESS;
}


static void
show_status(DBusGProxy* auditioner)
{
	p(1, "requesting status...");
	GError *error = NULL;
	char* status;
	dbus_g_proxy_call(auditioner, "getStatus", &error, G_TYPE_INVALID, G_TYPE_STRING, &status, G_TYPE_INVALID);
	if(error){
		p(0, "error: %s", error->message);
		g_error_free(error);
		return;
	}
	p(1, "got status.");
	p(0, "%s", status);
}


static void
log_handler(const gchar* log_domain, GLogLevelFlags log_level, const gchar* message, gpointer user_data)
{
	switch(log_level){
		case G_LOG_LEVEL_CRITICAL:
			printf("%s %s\n", ayyi_err, message);
			break;
		case G_LOG_LEVEL_WARNING:
			printf("%s %s\n", ayyi_warn, message);
			break;
		default:
			printf("log_handler(): level=%i %s\n", log_level, message);
			break;
	}
}


